﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VoiceElements.Common;
using System.Data;
using System.IO;
using System.Collections;

namespace Helper
{
    public class CsvParser
    {
        public static DataTable Parse(Log log, string data, bool headers)
        {
            return Parse(log, new StringReader(data), headers);
        }

        public static DataTable Parse(Log log, string data)
        {
            return Parse(log, new StringReader(data));
        }

        public static DataTable Parse(Log log, TextReader stream)
        {
            return Parse(log, stream, false);
        }

        public static DataTable Parse(Log log, TextReader stream, bool headers)
        {
            DataTable table = new DataTable();
            CsvStream csv = new CsvStream(stream);
            int headerCount = 0;
            string[] row = csv.GetNextRow();
            if (row == null)
                return null;
            if (headers)
            {
                headerCount = row.Length;
                foreach (string header in row)
                {
                    if (header != null && header.Length > 0 && !table.Columns.Contains(header))
                    {
                        table.Columns.Add(header, typeof(string));
                        //log.Write("AddedHeader1: {0}",header);
                    }
                    else
                    {
                        string s = GetNextColumnHeader(table);
                        table.Columns.Add(s, typeof(string));
                        //log.Write("AddedHeader2: {1}",header);
                    }
                }
                row = csv.GetNextRow();
                //for(int i=0;i<row.Length;i++) 
                //{
                //	log.Write("{0}: {1}",i,row[i].ToString());
                //}

            }
            while (row != null)
            {
                while (row.Length > table.Columns.Count)
                    table.Columns.Add(GetNextColumnHeader(table), typeof(string));
                if (headerCount > 0 && row.Length != headerCount)
                {
                    row = csv.GetNextRow();
                    return null;
                    //continue;
                }
                //log.Write("{0} {1}",row.Length,row[0].ToString());
                table.Rows.Add(row);
                row = csv.GetNextRow();
                //for(int i=0;i<row.Length;i++) 
                //{
                //	log.Write("{0}: {1}",i,row[i].ToString());
                //}
            }
            return table;
        }

        private static string GetNextColumnHeader(DataTable table)
        {
            int c = 1;
            while (true)
            {
                string h = "Column" + c++;
                if (!table.Columns.Contains(h))
                    return h;
            }
        }


        private class CsvStream
        {
            private TextReader stream;
            private bool delimeterFound;
            private char delimeter;

            public CsvStream(TextReader s)
            {
                stream = s;
                delimeterFound = false;
            }

            public string[] GetNextRow()
            {
                ArrayList row = new ArrayList();
                while (true)
                {
                    string item = GetNextItem();
                    if (item == null)
                        return row.Count == 0 ? null : (string[])row.ToArray(typeof(string));
                    row.Add(item);
                }
            }

            private bool EOS = false;
            private bool EOL = false;

            private string GetNextItem()
            {
                if (EOL)
                {
                    // previous item was last in line, start new line
                    EOL = false;
                    return null;
                }
                bool quoted = false;
                bool predata = true;
                bool postdata = false;
                StringBuilder item = new StringBuilder();

                while (true)
                {
                    char c = GetNextChar(true);
                    if (EOS)
                        return item.Length > 0 ? item.ToString() : null;
                    if (!delimeterFound)
                    {
                        if (c == ',')
                        {
                            delimeter = ',';
                            delimeterFound = true;
                        }
                        if (c == '\t')
                        {
                            delimeter = '\t';
                            delimeterFound = true;
                        }
                    }
                    if ((postdata || !quoted) && c == delimeter)
                        // end of item, return
                        return item.ToString();

                    if ((predata || postdata || !quoted) && (c == '\x0A' || c == '\x0D'))
                    {
                        // we are at the end of the line, eat newline characters and exit
                        EOL = true;
                        if (c == '\x0D' && GetNextChar(false) == '\x0A')
                            // new line sequence is 0D0A
                            GetNextChar(true);
                        return item.ToString();
                    }

                    if (predata && c == ' ')
                        // whitespace preceeding data, discard
                        continue;

                    if (predata && c == '"')
                    {
                        // quoted data is starting
                        quoted = true;
                        predata = false;
                        continue;
                    }

                    if (predata)
                    {
                        // data is starting without quotes
                        predata = false;
                        item.Append(c);
                        continue;
                    }

                    if (c == '"' && quoted)
                    {
                        if (GetNextChar(false) == '"')
                            // double quotes within quoted string means add a quote       
                            item.Append(GetNextChar(true));
                        else
                            // end-quote reached
                            postdata = true;
                        continue;
                    }

                    // all cases covered, character must be data
                    item.Append(c);
                }
            }

            private char[] buffer = new char[4096];
            private int pos = 0;
            private int length = 0;

            private char GetNextChar(bool eat)
            {
                if (pos >= length)
                {
                    length = stream.ReadBlock(buffer, 0, buffer.Length);
                    if (length == 0)
                    {
                        EOS = true;
                        return '\0';
                    }
                    pos = 0;
                }
                if (eat)
                    return buffer[pos++];
                else
                    return buffer[pos];
            }
        }
    }
}
